from __future__ import annotations

from typing import TYPE_CHECKING

from env_config.constants import Undefined

if TYPE_CHECKING:
    from env_config import Environment


__all__ = [
    "DjangoEnvConfigError",
    "MissingEnvValueError",
    "MissingExtraDependencyError",
]


class DjangoEnvConfigError(Exception):
    """Base class for all Django Environment Config errors."""


class MissingEnvValueError(DjangoEnvConfigError):
    """Error raised when a value is not found in the .env file, and a default is not provided."""

    def __init__(self, *, name: str, env: type[Environment]) -> None:
        msg = f"Value {name!r} in environment {env.__name__!r}"
        if env.dotenv_path is not Undefined:
            msg += " not defined in the .env file and value does not have a default"
        else:
            msg += " needs a default value since environment does not define a `dotenv_path`"

        super().__init__(msg)


class MissingExtraDependencyError(DjangoEnvConfigError):
    """Base class for all Django Environment Config errors."""



from __future__ import annotations

import sys
from typing import Any, Callable, Generator, Generic, Mapping, ParamSpec, Sequence, TypedDict, TypeVar

if sys.version_info < (3, 12):  # pragma: no cover
    from typing_extensions import Unpack
else:
    from typing import Unpack

__all__ = [
    "Any",
    "CacheConfig",
    "Callable",
    "DBConfig",
    "DBConfigExtra",
    "Generator",
    "Generic",
    "Mapping",
    "ParamSpec",
    "Sequence",
    "TypeVar",
    "Unpack",
]


class DBConfig(TypedDict, total=False):
    ATOMIC_REQUESTS: bool
    AUTOCOMMIT: bool
    CONN_MAX_AGE: int | None
    CONN_HEALTH_CHECKS: bool
    DISABLE_SERVER_SIDE_CURSORS: bool
    ENGINE: str
    HOST: str
    NAME: str
    OPTIONS: dict[str, Any] | None
    PASSWORD: str
    PORT: str | int
    TEST: dict[str, Any]
    TIME_ZONE: str
    USER: str


class DBConfigExtra(TypedDict, total=False):
    engine: str | None
    conn_max_age: int | None
    conn_health_checks: bool
    ssl_require: bool
    test_options: dict | None


class CacheConfig(TypedDict, total=False):
    BACKEND: str
    KEY_FUNCTION: str
    KEY_PREFIX: str
    LOCATION: str
    OPTIONS: dict[str, Any]
    TIMEOUT: int
    VERSION: int



from __future__ import annotations

import inspect
import os
from typing import TYPE_CHECKING

from django.utils.functional import classproperty
from dotenv import dotenv_values
from dotenv.main import find_dotenv

from .constants import ENV_NAME, Undefined

if TYPE_CHECKING:
    from dotenv.main import StrPath

    from .typing import Any

__all__ = [
    "Environment",
]


class Environment:
    """
    Configures a single environment.

    Creation of a subclasses can be customized by setting keyword arguments on the class.
    See `__init_subclass__` for more info.

    >>> class MyEnvironment(Environment, use_environ=True):
    >>>     pass
    """

    def __init_subclass__(
        cls,
        *,
        dotenv_path: StrPath | None | Undefined = Undefined,
        use_environ: bool = False,
    ) -> None:
        """
        When a subclass of environment is created, try to immediately load the settings
        and set them in the module globals where the environment is defined, if the environment matches
        what has been selected with the `DJANGO_SETTINGS_ENVIRONMENT` environment variable.

        :param dotenv_path: The path to the `.env` file to load. If set to `None`, the `.env` file will not be loaded.
                            By default, `python-dotenv` will try to find the `.env` file automatically.
        :param use_environ: If set to `True`, use environment variables instead of using a `.env` file.
        """
        env: str | None = os.environ.get(ENV_NAME)
        if env is None:  # pragma: no cover
            msg = f"Environment variable {ENV_NAME!r} must be set before subclassing 'Environment'"
            raise ValueError(msg)

        # If the environment does not match, do not load the environment or even the `.env` file.
        if cls.__name__.casefold() != env.casefold():
            return

        # If set to `None` explicitly, or using environment, do not load a `.env` file.
        if dotenv_path is None or use_environ:
            dotenv_path = Undefined

        # If not given, set it to `None` so the `dotenv.main.find_dotenv`
        # will try to find the `.env` file automatically.
        elif dotenv_path is Undefined:
            dotenv_path = None

        dotenv: dict[str, str] | Undefined
        if use_environ:
            dotenv = os.environ.copy()
        elif dotenv_path is not Undefined:
            dotenv = cls.load_dotenv(dotenv_path=dotenv_path)
        else:
            dotenv = Undefined

        # Do name mangling to avoid overriding the attribute from a parent class.
        # This way, we can have multiple environments with different `.env` files,
        # and allow using values from a parent `.env` file as defaults (if desired).
        setattr(cls, f"_{cls.__name__}__dotenv", dotenv)
        setattr(cls, f"_{cls.__name__}__dotenv_path", dotenv_path)

        cls.pre_setup()
        cls.setup(stack_level=2)
        cls.post_setup()

    @staticmethod
    def load_dotenv(*, dotenv_path: StrPath | None = None) -> dict[str, str]:  # pragma: no cover
        """Load the `.env` file and return the values."""
        if dotenv_path is None:
            dotenv_path = find_dotenv(raise_error_if_not_found=True, usecwd=True)
        return dotenv_values(dotenv_path=dotenv_path)

    @classmethod
    def pre_setup(cls) -> None:
        """
        Hook for doing additional setup before the settings have been loaded,
        but after the `.env` file has been loaded.
        """

    @classmethod
    def setup(cls, *, stack_level: int = 1) -> None:
        """Load settings and set them in the module globals where the environment is defined."""
        settings = cls.load_settings()
        stack = inspect.stack()
        module_globals: dict[str, Any] = stack[stack_level].frame.f_globals
        module_globals.update(**settings)

    @classmethod
    def post_setup(cls) -> None:
        """Hook for doing additional setup after the settings have been loaded."""

    @classmethod
    def load_settings(cls) -> dict[str, Any]:
        """Load the settings from the environment, validating and returning them."""
        return {name: getattr(cls, name) for name in dir(cls) if name.isupper() and not name.startswith("_")}

    @classproperty
    def dotenv(cls) -> dict[str, str] | Undefined:  # noqa: N805
        return getattr(cls, f"_{cls.__name__}__dotenv", Undefined)

    @classproperty
    def dotenv_path(cls) -> str | None | Undefined:  # noqa: N805
        return getattr(cls, f"_{cls.__name__}__dotenv_path", Undefined)



from __future__ import annotations

import os

import pytest

ENV_NAME = "DJANGO_SETTINGS_ENVIRONMENT"


@pytest.hookimpl()
def pytest_addoption(parser: pytest.Parser) -> None:
    # Adds the env variable as a recognized configuration option.
    parser.addini(ENV_NAME, "`django-environment-config` class to use by pytest-django.")


# Must use `tryfirst` to set the environment variable before Django is loaded by other plugins.
@pytest.hookimpl(tryfirst=True)
def pytest_load_initial_conftests(early_config: pytest.Config, parser: pytest.Parser, args: list[str]) -> None:
    # Sets the environment variable to the value of the `DJANGO_SETTINGS_ENVIRONMENT` option (if it exists).
    value: str = early_config.getini(ENV_NAME)
    if value:
        os.environ.setdefault(ENV_NAME, value)



from __future__ import annotations

__all__ = [
    "ENV_NAME",
    "Undefined",
]


class Undefined:  # pragma: no cover
    """Value to represent an undefined value."""

    def __repr__(self) -> str:
        return self.__class__.__name__

    def __bool__(self) -> bool:
        return False


Undefined = Undefined()

ENV_NAME = "DJANGO_SETTINGS_ENVIRONMENT"



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from collections import defaultdict
from decimal import Decimal
from pathlib import Path
from typing import TYPE_CHECKING

from django.utils.module_loading import import_string

from .constants import Undefined
from .errors import MissingEnvValueError, MissingExtraDependencyError
from .typing import Any, CacheConfig, DBConfig, DBConfigExtra, Generator, Generic, Mapping, Sequence, TypeVar, Unpack

if TYPE_CHECKING:
    from .base import Environment


__all__ = [
    "BooleanValue",
    "CacheURLValue",
    "DatabaseURLValue",
    "DecimalValue",
    "DictValue",
    "EmailValue",
    "FloatValue",
    "IPValue",
    "ImportStringValue",
    "IntegerValue",
    "JsonValue",
    "ListValue",
    "MappingValue",
    "PathValue",
    "PositiveIntegerValue",
    "RegexValue",
    "SequenceValue",
    "SetValue",
    "StringValue",
    "TupleValue",
    "URLValue",
    "Value",
]


T = TypeVar("T")


class Value(ABC, Generic[T]):
    def __init__(
        self,
        *,
        default: T | None = Undefined,
        env_name: str | None | Undefined = Undefined,
    ) -> None:
        """
        Value descriptor for an environment variable.

        :param default: The default value to use if the environment variable is not set.
        :param env_name: The name of the environment variable to use. If not given, the name of the field is used.
                         Set this to `None` to skip loading the value from the environment.
        """
        self.default: T | None = default
        self.name: str = env_name
        self.skip_env: bool = env_name is None

        # Use a map to store the value per environment so that we can have
        # different values for environments what inherit from each other.
        self.value_by_environment: dict[type[Environment], Any] = defaultdict(lambda: Undefined)
        super().__init__()

    def __set_name__(self, env: type[Environment], name: str) -> None:
        """Called after the owner Environment-class is created with this field as a class attribute."""
        if self.name in (Undefined, None):
            self.name = name

    def __get__(self, _: Environment | None, env: type[Environment]) -> T:
        """Called when accessing the field on the class or an instance of the class."""
        if self.value_by_environment[env] is not Undefined:
            return self.value_by_environment[env]

        self.value_by_environment[env] = self.get_for_environment(env)
        return self.value_by_environment[env]

    def get_for_environment(self, env: type[Environment]) -> T:
        value = self.default if env.dotenv is Undefined or self.skip_env else env.dotenv.get(self.name, self.default)
        if value is Undefined:
            raise MissingEnvValueError(name=self.name, env=env)

        if value is None:
            return None

        return self.convert(value)

    @abstractmethod
    def convert(self, value: str | T) -> T:  # pragma: no cover
        """Convert the given value into the proper representation."""
        raise NotImplementedError


class StringValue(Value[str]):
    """Parses env variables into a string value."""

    def convert(self, value: str) -> str:
        return value


class BooleanValue(Value[bool]):
    """Parses env variables into a boolean value."""

    def convert(self, value: str | bool) -> bool:  # noqa: FBT001
        if isinstance(value, bool):
            return value
        normalized_value = value.strip().lower()
        if normalized_value in ("yes", "y", "true", "1"):
            return True
        if normalized_value in ("no", "n", "false", "0", ""):
            return False
        msg = f"Cannot interpret {value!r} as a boolean value"
        raise ValueError(msg)


class IntegerValue(Value[int]):
    """Parses env variables into an integer value."""

    def convert(self, value: str | int) -> int:
        return int(value)


class PositiveIntegerValue(IntegerValue):
    """Parses env variables into an integer value, and validates that the value is positive."""

    def convert(self, value: str | int) -> int:
        val = super().convert(value)
        if val < 0:
            msg = f"Value must be positive, got {val}"
            raise ValueError(msg)
        return val


class FloatValue(Value[float]):
    """Parses env variables into a float value."""

    def convert(self, value: str | float) -> float:
        return float(value)


class DecimalValue(Value[Decimal], Decimal):
    """Parses env variables into a Decimal value."""

    def convert(self, value: str | Decimal) -> Decimal:
        return Decimal(value)


class ImportStringValue(Value[str]):
    """Parses env variables into a string value, and validates that the value is an importable string."""

    def convert(self, value: str) -> str:
        import_string(value)
        return value


class SequenceValue(Value, ABC, Generic[T]):
    def __init__(
        self,
        child: Value[T] | None = None,
        *,
        default: Sequence[T] | None = Undefined,
        env_name: str | None | Undefined = Undefined,
        delimiter: str = ",",
    ) -> None:
        self.child = child or StringValue()
        self.delimiter = delimiter
        super().__init__(default=default, env_name=env_name)

    def iterate(self, value: str | Sequence[Any]) -> Generator[T, None, None]:
        seq = value.split(self.delimiter) if isinstance(value, str) else value
        for item in seq:
            if not item:
                continue

            yield self.child.convert(item.strip())


class ListValue(SequenceValue):
    """Parses env variables like `item1,item2,item3` into a list."""

    def convert(self, value: str | list[Any]) -> list[Any]:
        return list(self.iterate(value))


class TupleValue(SequenceValue):
    """Parses env variables like `item1,item2,item3` into a tuple."""

    def convert(self, value: str | tuple[Any, ...]) -> tuple[Any, ...]:
        return tuple(self.iterate(value))


class SetValue(SequenceValue):
    """Parses env variables like `item1,item2,item3` into a set."""

    def convert(self, value: str | set[str]) -> set[Any]:
        return set(self.iterate(value))


class MappingValue(Value, ABC, Generic[T]):
    def __init__(  # noqa: PLR0913
        self,
        child: Value[T] | None = None,
        *,
        default: Mapping[str, T] | None = Undefined,
        env_name: str | None | Undefined = Undefined,
        kv_delimiter: str = "=",
        item_delimiter: str = ";",
    ) -> None:
        self.child = child or StringValue()
        self.kv_delimiter = kv_delimiter
        self.item_delimiter = item_delimiter
        super().__init__(default=default, env_name=env_name)

    def iterate(self, value: str | Mapping[str, Any]) -> Generator[tuple[str, Any], None, None]:
        seq = value.split(self.item_delimiter) if isinstance(value, str) else value.items()

        for item in seq:
            if not item:
                continue

            if isinstance(item, str):
                kv = item.split(self.kv_delimiter, 1)
                if len(kv) != 2:  # noqa: PLR2004
                    msg = f"Cannot split key-value pair from {item!r}"
                    raise ValueError(msg)
            else:
                kv = item

            yield kv[0].strip(), self.child.convert(kv[1].strip())


class DictValue(MappingValue):
    """Parses env variables like `key1=value1;key2=value2` into a dict."""

    def convert(self, value: str | dict[str, Any]) -> dict[str, Any]:
        return dict(self.iterate(value))


class JsonValue(Value[dict | list]):
    """Parses env variables from a json string to a python list or dict."""

    def convert(self, value: str | list | dict) -> list | dict:
        if isinstance(value, (list, dict)):
            return value
        return json.loads(value)


class EmailValue(StringValue):
    """Parses env variables into a string value, and validates that it's a valid email address."""

    def convert(self, value: str) -> str:
        from django.core.validators import validate_email

        validate_email(value)
        return value


class URLValue(StringValue):
    """Parses env variables into a string value, and validates that it's a valid URL."""

    def convert(self, value: str) -> str:
        from django.core.validators import URLValidator

        URLValidator()(value)
        return value


class IPValue(StringValue):
    """Parses env variables into a string value, and validates that it's a valid IP address."""

    def convert(self, value: str) -> str:
        from django.core.validators import validate_ipv46_address

        validate_ipv46_address(value)
        return value


class RegexValue(StringValue):
    """Parses env variables into a string value, and validates that it matches the given regex."""

    def __init__(
        self,
        *,
        regex: str,
        default: str | None = Undefined,
        env_name: str | None | Undefined = Undefined,
    ) -> None:
        self.regex = regex
        super().__init__(default=default, env_name=env_name)

    def convert(self, value: str) -> str:
        from django.core.validators import RegexValidator

        RegexValidator(regex=self.regex)(value)
        return value


class PathValue(StringValue):
    """Parses env variable into a string value, and can optionally validate that the path exists."""

    def __init__(  # noqa: PLR0913
        self,
        *,
        default: str | None = Undefined,
        env_name: str | None | Undefined = Undefined,
        check_exists: bool = True,
        create_if_missing: bool = False,
        mode: int = 0o777,
    ) -> None:
        self.check_exists = check_exists
        self.create_if_missing = create_if_missing
        self.mode = mode
        super().__init__(default=default, env_name=env_name)

    def convert(self, value: str) -> str:
        path = Path(value).absolute()
        if self.create_if_missing:
            path.mkdir(mode=self.mode, parents=True, exist_ok=True)
        if self.check_exists and not path.exists():
            msg = f"Path '{path}' does not exist"
            raise ValueError(msg)

        return str(path)


class DatabaseURLValue(Value[DBConfig | str]):
    """Load a database configuration from a URL."""

    def __init__(
        self,
        *,
        db_alias: str = "default",
        default: DBConfig | str = Undefined,
        env_name: str = "DATABASE_URL",
        **params: Unpack[DBConfigExtra],
    ) -> None:
        self.db_alias = db_alias
        self.params = params
        super().__init__(default=default, env_name=env_name)

    def convert(self, value: str | DBConfig) -> dict[str, DBConfig]:
        if not isinstance(value, str):
            return {self.db_alias: value}

        try:
            from dj_database_url import parse
        except ImportError as error:  # pragma: no cover
            msg = (
                "You must install the 'db' extra dependency "
                "(e.g., `pip install django-environment-config[db]`) "
                "to use the DatabaseURLValue."
            )
            raise MissingExtraDependencyError(msg) from error

        config = parse(value, **self.params)
        return {self.db_alias: config}


class CacheURLValue(Value[CacheConfig | str]):
    """Load a cache configuration from a URL."""

    def __init__(
        self,
        *,
        cache_alias: str = "default",
        default: CacheConfig | str = Undefined,
        env_name: str = "CACHE_URL",
    ) -> None:
        self.cache_alias = cache_alias
        super().__init__(default=default, env_name=env_name)

    def convert(self, value: str | CacheConfig) -> dict[str, CacheConfig]:
        if not isinstance(value, str):
            return {self.cache_alias: value}

        try:
            from django_cache_url import parse
        except ImportError as error:  # pragma: no cover
            msg = (
                "You must install the 'cache' extra dependency "
                "(e.g., `pip install django-environment-config[cache]`) "
                "to use the CacheURLValue."
            )
            raise MissingExtraDependencyError(msg) from error

        config = parse(value)
        return {self.cache_alias: config}


class ParentValue(Value[T]):
    """Parses value from the parent environment, or the default value if parent doesn't have the value."""

    def __init__(
        self,
        child: Value[T] | None = None,
        *,
        default: T | None,
        env_name: str | None | Undefined = Undefined,
        check_limit: int | None = None,
    ) -> None:
        self.child = child or StringValue()
        self.env: type[Environment] | None = None
        self.check_limit = check_limit if check_limit is None else check_limit + 1
        super().__init__(default=default, env_name=env_name)

    def get_for_environment(self, env: type[Environment]) -> None:
        for parent in env.mro()[slice(1, self.check_limit)]:
            value: Any = getattr(parent, self.name, Undefined)
            if value is not Undefined:
                return self.convert(value)
        return self.convert(self.default)

    def convert(self, value: str | T) -> T:
        return self.child.convert(value)
